unit Ribbon;
{
Editable Ribbon Unit, version 1.1, 15 feb 1994
----------------------------------------------

     CopyRight  : 1994, N.T.O.B., the Netherlands
        Author  : Hans de Bue
     CompuServ  : 100034, 2645
Initial release : 10 feb 1994
}


interface

uses
  BWCC,
  Objects,
  ODialogs,
  OWindows,
  Strings,
  WinProcs,
  WinTypes;

{$R RIBBON.RES }

{ Predefined buttons in RIBBON.RES

201  : New
202  : Open
203  : Save
204  : Print
205  : Cut
206  : Copy
207  : Paste
208  : Help
209  : Help with arrow
210  : Sum sign
}

const
  rbn_EnableEdit     = TRUE;
  rbn_DisableEdit    = FALSE;
  btn_InstallOn      = TRUE;
  btn_InstallOff     = FALSE;

type
  PRibbon            = ^TRibbon;
  TRibbon            = object( TWindow )
    Height           : integer;
    EnableEdit       : boolean;
    AllButtons       : PCollection;
    Editor           : PWindow;

    constructor Init( aParent: PWindowsObject; EditFlag: Boolean );
    function    GetClassName: PChar; virtual;
    procedure   GetWindowClass( var aWndClass: TWndClass ); virtual;
    procedure   Resize( Msg: TMessage );
    procedure   Edit;
    procedure   Visible;
    procedure   Hide;
    procedure   ReSizeParent;

    procedure   AddButton( aBtnID, aCmdID: word;
                           aDescription: PChar; InstallFlag: Boolean );
    procedure   AddBtn( aBtnID, aCmdID: word; aName: PChar );
    procedure   AddSeparator;
    procedure   AddSep;
    function    GetNextXCoord: integer;

    procedure   EnableBtn( aCmdID, State: word );
    procedure   ShowButtons;

    procedure   EditModeOn;
    procedure   EditModeOff;
    procedure   EditMode( Flag: Boolean );
    procedure   EnableAll;
    function    Install( aBtnID: integer; State: Boolean ): boolean;

    procedure   WMLButtonDblClk( var Msg: TMessage );
      virtual wm_First + wm_LButtonDblClk;
  end; { TRibbon }

implementation

{=========================================================================}
{== Local definitions                                                   ==}
{=========================================================================}

type
  PRibbonBtn         = ^TRibbonBtn;
  TRibbonBtn         = object( TButton )
    BtnID            : word;
    Command          : word;
    Name             : PChar;
    EditMode         : boolean;
    PrevState        : boolean;

    constructor Init( aParent: PWindowsObject; aName: PChar;
                      aBtnID, aCommand, NewX: word );
    function    InClientRect( Window: HWnd; Psn: LongInt ): boolean;
    procedure   WMLButtonUp( var Msg: TMessage );
      virtual wm_First + wm_LButtonUp;
  end; { TRibbonBtn }


  PEditBtn           = ^TEditBtn;
  TEditBtn           = object( TButton )
    BtnID            : word;
    CmdID            : word;
    Description      : PChar;

    constructor Init( aParent: PWindowsObject; aBtnID, aCmdID: word;
                      aDescription: PChar; X, Y: integer);
    procedure   WMLButtonUp( var Msg: TMessage );
      virtual wm_First + wm_LButtonUp;
    procedure   WMNCHitTest( var Msg: TMessage );
      virtual wm_First + wm_NCHitTest;
  end; { TEditBtn }


  PSeparator         = ^TSeparator;
  TSeparator         = object( TWindow )
    constructor Init( aParent: PWindowsobject; aPosition: integer );
  end; { TSeparator }


  PBtnRec            = ^TBtnRec;
  TBtnRec            = object( TObject )
    BtnID            : word;
    CmdID            : word;
    Description      : array[0..64] of char;
    Install          : boolean;

    constructor Init( aBtnID, aCmdID: word; aDescription: PChar;
                      InstallFlag: boolean );
    constructor Load( var S: TDosStream );
    procedure   Store( var S: TDosStream );
  end; { TBtnRec }


  PEditRibbon        = ^TEditRibbon;
  TEditRibbon        = object( TWindow )
    Ribbon           : PRibbon;
    Buttons          : PCollection;
    BtnOK            : PButton;
    BtnDescription   : PStatic;

    constructor Init( aRibbon: PRibbon );
    procedure   SetupWindow; virtual;
    function    GetClassName: PChar; virtual;
    procedure   EnableBtn( BtnID: word );
    procedure   IDBtnOK( var Msg: TMessage );
      virtual id_First + id_OK;
  end; { TEditRibbon }


const
  id_BtnDescription  = 999; { '999' should not be used for a ribbon-button }
  id_Separator       = 0;   { same applies for '0' }

  RBtnRec          : TStreamRec = (
    ObjType        : 1010;
    VmtLink        : Ofs( TypeOf( TBtnRec )^ );
    Load           : @TBtnRec.Load;
    Store          : @TBtnRec.Store
  ); { RWpt }

{=========================================================================}
{== Implementation of TRibbonBtn                                        ==}
{=========================================================================}

constructor TRibbonBtn.Init( aParent: PWIndowsObject; aName: PChar;
                             aBtnID, aCommand, NewX: word );

begin
  inherited Init( aParent, aBtnID, nil, NewX, 2, 24, 22, FALSE );
  BtnID := aBtnID;
  Command := aCommand;
  Name := aName;
  EditMode := FALSE;
end; { TRibbonBtn.Init }

{=========================================================================}

function TRibbonBtn.InClientRect( Window: HWnd; Psn: LongInt ): boolean;

var
  Rect             : TRect;
  X, Y             : word;

begin
  InClientRect := FALSE;
  GetClientRect( Window, Rect );
  X := LoWord( Psn );
  Y := HiWord( Psn );
  with Rect do
    if (X>Left) and (X<Right) and (Y>Top) and (Y<Bottom) then
      InClientRect := TRUE;
end; { TRibbonBtn.InClientRect }

{=========================================================================}

procedure TRibbonBtn.WMLButtonUp( var Msg: TMessage );

begin
  if not(EditMode) then
    begin
      if InClientRect( HWindow, Msg.lParam ) then
        begin
          SetFocus(Application^.MainWindow^.HWindow);
          DefWndProc( Msg );
          SendMessage( Application^.MainWindow^.HWindow, wm_Command, Command, 0 );
        end
    end { if }
  else
    with PRibbon( Parent )^ do
      begin
        if Install( BtnID, btn_InstallOff ) then
          begin
            ShowButtons;
            EditModeOn;
          end; { if }
        DefWndProc( Msg );
      end; { with }
end; { TRibbonBtn.WMLButtonUp }

{=========================================================================}
{== Implementation of TEditBtn                                          ==}
{=========================================================================}

constructor TEditBtn.Init( aParent: PWindowsObject; aBtnID, aCmdID: word;
                           aDescription: PChar; X, Y: integer);

begin
  inherited Init( aParent, aBtnID, nil, X, Y, 24, 22, FALSE );
  BtnID := aBtnID;
  CmdID := aCmdID;
  Description := aDescription;
end; { TEditBtn.Init }

{=========================================================================}

procedure TEditBtn.WMLButtonUp( var Msg: TMessage );

begin
  with PEditRibbon(Parent)^.Ribbon^ do
    begin
      if Install( BtnID, btn_InstallOn ) then
        begin
          ShowButtons;
          EditModeOn;
        end;
    end; { with }
  DefWndProc( Msg );
end; { TEditBtn.WMLButtonUp }

{=========================================================================}

procedure TEditBtn.WMNCHitTest( var Msg: TMessage );

var
  Bfr              : array[0..64] of char;
begin
  PEditRibbon( Parent )^.BtnDescription^.GetText( Bfr, SizeOf( Bfr ));
  if StrComp( Bfr, Description ) <> 0 then
    PEditRibbon( Parent )^.BtnDescription^.SetText( Description );
end; { TEditBtn.WMNCHitTest }

{=========================================================================}
{== Implementation of TSeparator                                        ==}
{=========================================================================}

constructor TSeparator.Init( aParent: PWindowsObject; aPosition: integer );

begin
  inherited Init( aParent, nil );
  with Attr do
    begin
      X := aPosition;
      Y := 2;
      W := 10;
      H := 5;
      Style := ws_Child;
    end; { with }
end; { TSeparator.Init }

{=========================================================================}
{== Implementation of TBtnRec                                           ==}
{=========================================================================}

constructor TBtnRec.Init( aBtnID, aCmdID: word; aDescription: PChar;
                          InstallFlag: boolean );

begin
  inherited Init;
  BtnID := aBtnID;
  CmdID := aCmdID;
  StrCopy( Description, aDescription );
  Install := InstallFlag;
end; { TBtnRec.Init }

{ ======================================================================= }

constructor TBtnRec.Load( var S: TDosStream );

begin
  S.Read( BtnID, SizeOf( BtnID ));
  S.Read( CmdID, SizeOf( CmdID ));
  S.Read( Description, SizeOf( Description ));
  S.Read( Install, SizeOf( Install ));
end; { TBtnRec.Load }

{ ======================================================================= }

procedure TBtnRec.Store( var S: TDosStream );

begin
  S.Write( BtnID, SizeOf( BtnID ));
  S.Write( CmdID, SizeOf( CmdID ));
  S.Write( Description, SizeOf( Description ));
  S.Write( Install, SizeOf( Install ));
end; { TBtnRec.Store }

{=========================================================================}
{== Implementation of TEditRibbon                                       ==}
{=========================================================================}

constructor TEditRibbon.Init( aRibbon: PRibbon );

var
  NrOfRows         : integer;

begin
  inherited Init( Application^.MainWindow, 'Edit Ribbon' );
  Ribbon := aRibbon;
  Buttons := Ribbon^.AllButtons;
  with Attr do
    begin
      SetFlags( wb_MDIChild, False );
      Style := ws_Overlapped or ws_Visible;
      X := 200;
      Y := 100;
      W := 270;
      NrOfRows := Buttons^.Count div 10;
      H := 150 + NrOfRows*25;
    end; { with }
  BtnOK := New( PButton, Init( @Self, 1, nil,
                100 , 75 + NrOfRows*25, 40, 15, TRUE ));
  BtnDescription := New( PStatic, Init( @Self, id_BtnDescription, 'Click the buttons...',
                         5, 30 + NrOfRows*25, 265, 40, 64 ));
end; { TEditRibbon.Init }

{=========================================================================}

procedure TEditRibbon.SetupWindow;

var
  Count, X, Y      : integer;
  BtnRec           : PBtnRec;
  Btn              : PButton;
  BtnNr            : integer;

begin
  inherited SetupWindow;
  X := 5;
  BtnNr := 0;
  for Count := 0 to Buttons^.Count-1 do
    begin
      BtnRec := Buttons^.At( Count );
      if StrComp( BtnRec^.Description, 'SEPARATOR' ) <> 0 then
        begin
          Y := 5 + (BtnNr div 10)*25;
          if X > 240 then
            X := 5;
          Btn := New( PEditBtn, Init( @Self, BtnRec^.BtnID, BtnRec^.CmdID,
                      BtnRec^.Description, X, Y ));
          Application^.MakeWindow( Btn );
          Inc( X, 26 );
          Inc( BtnNr );
      end; { if }
    end; { for }

  { put all buttons in the ribbon in edit-mode }
  Ribbon^.EnableAll;
  Ribbon^.EditModeOn;
end; { TEditRibbon.SetupWindow }

{=========================================================================}

function TEditRibbon.GetClassName: PChar;

begin
  GetClassName := 'EditRibbon';
end; { TEditRibbon.GetClassName }

{=========================================================================}

procedure TEditRibbon.EnableBtn( BtnID: word );

var
  FoundObject      : PWindowsObject;

  function Matches( anObject: PControl ): boolean; far;
  begin
    Matches := FALSE;
    if anObject^.Attr.ID = BtnID then
      Matches := TRUE;
  end; { Matches }

begin
  FoundObject := FirstThat( @Matches );
  if FoundObject <> nil then
    FoundObject^.Enable;
end; { TEditRibbon.EnableBtn }

{=========================================================================}

procedure TEditRibbon.IDBtnOK( var Msg: TMessage );

begin
  { put all buttons in the ribbon in working-mode }
  Ribbon^.EditModeOff;
  Ribbon^.Editor := nil;
  Done;
  { Sends a wm_SetFocus to the main window. Using a handler over there
    sets the ribbon-buttons in their appropriate state: enabled or grayed }
  SetFocus(Application^.MainWindow^.HWindow);
end; { TEditRibbon,IDBtnOK }

{=========================================================================}
{== Implementation of TRibbon                                           ==}
{=========================================================================}

constructor TRibbon.Init(AParent: PWindowsObject; EditFlag: Boolean);

begin
  inherited Init(AParent, '');
  SetFlags( wb_MDIChild, False );
  Attr.Style := ws_Border or ws_Child or ws_Visible;
  Height := 30;
  EnableEdit := EditFlag;
  AllButtons := New( PCollection, Init( 1, 1 ));
  Editor := nil;
end; { TRibbon.Init }

{=========================================================================}

function TRibbon.GetClassName: PChar;

begin
  GetClassName := 'Ribbon';
end; { TRibbon.GetClassName }

{=========================================================================}

procedure TRibbon.GetWindowClass( var aWndClass: TWndClass );

begin
  inherited GetWindowClass( aWndClass );
   AWndClass.hbrBackground := Gray_Brush;
   AWndClass.Style := AWndClass.Style or cs_DblClks;
end; { TRibbon.GetWindowClass }

{=========================================================================}

procedure TRibbon.Resize( Msg: TMessage );

begin
  if (HWindow <> 0) then
    if IsWindowVisible( HWindow ) then
      begin
        MoveWindow( HWindow, -1, -1, Msg.lParamLo+2, Height, TRUE );
        DefWndProc( Msg );
      end;
end; { TRibbon.Resize }

{=========================================================================}

procedure TRibbon.Edit;

begin
  if EnableEdit and (Editor=nil) and IsWindowVisible( HWindow )then
    begin
      Editor := New( PEditRibbon, Init( @Self ));
      Application^.MakeWindow( Editor );
    end; { if }
end; { TRibbon.Edit }

{=========================================================================}

procedure TRibbon.Visible;

begin
  Height := 30;
  ReSizeParent;
  Show( sw_ShowNormal );
end; { TRibbon.Visible }

{=========================================================================}

procedure TRibbon.Hide;

begin
  Show( sw_Hide );
  Height := 0;
  ReSizeParent;
end; { TRibbon.Hide }

{=========================================================================}

procedure TRibbon.ReSizeParent;

var
  Rect             : TRect;
  Msg              : TMessage;

begin
  GetClientRect(Parent^.HWindow, Rect );
  Msg.lParamLo := Rect.Right - Rect.Left;
  Msg.lParamHi := Rect.Bottom - Rect.Top;
  SendMessage(Parent^.HWindow, wm_Size, 0, Msg.lParam );
end; { TRibbon.ReSizeParent }

{=========================================================================}

procedure TRibbon.AddButton( aBtnID, aCmdID: word;
                             aDescription: PChar; InstallFlag: boolean );

var
  BtnRec           : PBtnRec;

begin
  BtnRec := New( PBtnRec, Init( aBtnID, aCmdID, aDescription, InstallFlag ));
  AllButtons^.Insert( BtnRec );
end; { TRibbon.AddButton }

{=========================================================================}

procedure TRibbon.AddBtn( aBtnID, aCmdID: word; aName: PChar );

var
  RibbonBtn        : PRibbonBtn;

begin
  RibbonBtn := New( PRibbonBtn, Init( @Self, aName, aBtnID, aCmdID,
                    GetNextXCoord ));
  Application^.MakeWindow( RibbonBtn );
end; { TRibbon.AddBtn }

{=========================================================================}

procedure TRibbon.AddSeparator;

var
  BtnRec           : PBtnRec;

begin
  BtnRec := New( PBtnRec, Init( id_Separator, 0, 'SEPARATOR', TRUE ));
  AllButtons^.Insert( BtnRec );
end; { TRibbon.AddSeparator }

{=========================================================================}

procedure TRibbon.AddSep;

var
  Separator        : PSeparator;

begin
  Separator := New( PSeparator, Init( @Self, GetNextXCoord ));
  Application^.MakeWindow( Separator );
end; { TRibbon.AddSep }

{=========================================================================}

function TRibbon.GetNextXCoord: integer;

var
  X                : integer;

  procedure GetPosition( Child: PWindow ); far;
  begin
    Inc( X, Child^.Attr.W  )
  end;

begin
  X := 2;
  ForEach( @GetPosition );
  GetNextXCoord := X;
end; { TRibbon.GetNextXCoord }

{=========================================================================}

procedure TRibbon.EnableBtn( aCmdID, State: word );

var
  FoundBtn         : PWindowsObject;

  function Matches( Child: PWindow ): boolean; far;
  begin
    Matches := FALSE;
    if Child^.Attr.ID <> 0 then
      if PRibbonBtn( Child )^.Command = aCmdID then
        Matches := TRUE;
  end; { Matches }

begin
  FoundBtn := FirstThat( @Matches );
  if FoundBtn <> nil then
    case State of
      mf_Enabled : PRibbonBtn( FoundBtn )^.Enable;
      mf_Grayed  : PRibbonBtn( FoundBtn )^.Disable;
    end; { case }
end; { TRibbon.EnableBtn }

{=========================================================================}

procedure TRibbon.ShowButtons;

  procedure RemoveIt( Child: PWindow ); far;
  begin
    Child^.Done;
  end; { RemoveIt }

  procedure AddIt( BtnRec: PBtnRec ); far;
  begin
    with BtnRec^ do
      if Install then
        if BtnID = id_Separator then
          begin
            { check if previous child <> separator }
            if ChildList <> nil then
              if PWindow(ChildList)^.Attr.ID <> id_Separator then
                AddSep
          end { if }
        else
          AddBtn( BtnID, CmdID, Description );
  end; { AddIt }

begin
  { Clear childlist }
  ForEach( @RemoveIt );
  { Rebuild childlist }
  AllButtons^.ForEach( @AddIt );
end; { TRibbon.ShowButtons }

{=========================================================================}

procedure TRibbon.EditModeOn;

begin
  EditMode( TRUE );
end; { TRibbon.EditModeOn }

{=========================================================================}

procedure TRibbon.EditModeOff;

begin
  EditMode( FALSE );
end; { TRibbon.EditModeOff }

{=========================================================================}

procedure TRibbon.EditMode( Flag: Boolean);

  procedure SetEditMode( Child: PWindow ); far;
  begin
    if Child^.Attr.ID <> id_Separator then
      PRibbonBtn( Child )^.EditMode:= Flag ;
  end; { SetEditMode }

begin
  ForEach( @SetEditMode );
end; { TRibbon.EditMode }

{=========================================================================}

procedure TRibbon.EnableAll;

  procedure EnableChild( Child: PWindow ); far;
  begin
    Child^.Enable;
  end; { SetEditMode }

begin
  ForEach( @EnableChild );
end; { TRibbon.EnableAll }

{=========================================================================}

function TRibbon.Install( aBtnID: integer; State: Boolean ): boolean;

var
  FoundBtnRec      : PBtnRec;

  function Matches( BtnRec: PBtnRec ): boolean; far;
  begin
    Matches := FALSE;
    if BtnRec^.BtnID = aBtnID then
      Matches := TRUE;
  end; { Matches }

begin
  Install := FALSE;
  FoundBtnRec := AllButtons^.FirstThat( @Matches );
  if FoundBtnRec <> nil then
    if FoundBtnRec^.Install <> State then
      begin
        FoundBtnRec^.Install := State;
        Install := TRUE;
      end; { if }
end; { TRibbon.Install }

{=========================================================================}

procedure TRibbon.WMLButtonDblClk( var Msg: TMessage );

begin
  Edit;
end; { TRibbon.WMLButtonDblClk }

{=========================================================================}
{== Unit initialisation code                                            ==}
{=========================================================================}

begin
  RegisterType( RBtnRec );
  RegisterType( RCollection );
end. { unit }